/*-
 * fbaccess0.c
 *
 * A sample code for frame buffer access.
 * Kazutaka YOKOTA.
 */

#include <stdio.h>
#include <err.h>
#include <signal.h>
#include <setjmp.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fbio.h>
#include <sys/consio.h>

#define TRUE	1
#define FALSE	0

/* Our virtual terminal. */
static int vtfd;

/* Set to TRUE, if the program is to terminate. */
static int quit = FALSE;

/* The flag to indicate if we are running the main loop. */
static int in_main_loop = FALSE;

/* The frame buffer window. */
static u_char *window = MAP_FAILED;
static int win_size = 0;		/* window size */
static int buf_size = 0;		/* frame buffer size */
static int line_width;
static int bpp;				/* bits per pixel */
static int planes;			/* color planes */

/* The buffer for setjmp/longjmp. */
static jmp_buf loop;

/* Function prototypes */
static void draw_something(void);
static void sig_release(int);
static void sig_acquire(int);
static void sig_term(int);

int
main(int argc, char *argv[])
{
	video_adapter_info_t adp_info;
	video_info_t mode_info;
	struct vt_mode prev_vtmode;
	struct vt_mode vtmode;
	int adapter;
	int prev_mode;
	int mode;

	/* Use stdin. */
	vtfd = 0;

	/*
	 * Check if we have a video adapter. If we are running in a serial
	 * or network terminal, the following ioctl will fail.
	 */
	if (ioctl(vtfd, FBIO_ADAPTER, &adapter))	/* CONS_CURRENTADP */
		err(1, "the program must run in a vty");

	/* Desired video mode */
	mode = M_VESA_CG1024x768;

	/*
	 * Obtain video mode information.  If this ioctl fails, the video
	 * card does not support the specified mode.
	 */
	mode_info.vi_mode = mode;
	if (ioctl(vtfd, FBIO_MODEINFO, &mode_info))	/* CONS_MODEINFO */
		err(1, "video mode %d is not supported", mode);

	/* Set up VT switching mode */
	ioctl(vtfd, VT_GETMODE, &prev_vtmode);
	vtmode.mode = VT_PROCESS;
	vtmode.relsig = SIGUSR1;
	vtmode.acqsig = SIGUSR2;
	vtmode.frsig = SIGUSR2;		/* not used XXX */
	signal(SIGUSR1, sig_release);
	signal(SIGUSR2, sig_acquire);
	signal(SIGINT, sig_term);
	signal(SIGTERM, sig_term);
	ioctl(vtfd, VT_SETMODE, &vtmode);

	/* Wait until our vty is brought to foreground. */
	ioctl(vtfd, VT_WAITACTIVE, 0);

	/* Save the current video mode */
	ioctl(vtfd, FBIO_GETMODE, &prev_mode);	/* CONS_GET */

	/* Set our video mode */		/* SW_XXX */
	ioctl(vtfd, FBIO_SETMODE, &mode);

	/*
	 * If we have set up a text mode and want to have complete
	 * control over the screen, we need to prevent the console
	 * driver from automatically updating the vty.  We also want to
	 * disable the screen saver.  It may sound strange to tell
	 * the console driver that the video card is in the GRAPHICS mode,
	 * when in fact we are using a text mode.  But, this will
	 * precisely do what we want to do to the console driver.
	 *
	 * ioctl(vtfd, KDSETMODE, KD_GRAPHCIS);
	 *
	 * If we have set up a "real" graphics mode, the above call is
	 * not necessary.  The console driver is clever enought to know
	 * that it shouldn't update the vty anymore...
	 */

	/*
	 * If you want to have direct access to the video card registers,
	 * call the following ioctl.
	 *
	 * if (ioctl(vtfd, KDENABIO, 0)) {
	 *	ioctl(vtfd, FBIO_SETMODE, &prev_mode);
	 *	ioctl(vtfd, VT_SETMODE, &prev_vtmode);
	 *	exit(1);
	 * }
	 */

	/*
	 * Obtain the adapter state and information.  It is essential to
	 * call this ioctl to examine exactly how the video card is set up.
	 * Window size and buffer size parameters must be obtained by 
	 * FBIO_ADPINFO. The corresponding values obtained by FBIO_MODEINFO
	 * shouldn't be used. 
	 */
	ioctl(vtfd, FBIO_ADPINFO, &adp_info);	/* CONS_ADPINFO */
	win_size = adp_info.va_window_size;
	buf_size = adp_info.va_buffer_size;
	line_width = adp_info.va_line_width;

	/* The other information is available in mode_info.  */
	bpp = mode_info.vi_depth;
	planes = mode_info.vi_planes;

	/*
	 * If win_size < buf_size/planes, the video card is in the window 
	 * mode. The program must switch memory banks to access the desired
	 * part of the frame buffer. 
	 */

	/* Now everything is set... */

	/*
	 * Map the window/frame buffer to the process address space.
	 */
	window = mmap(0, win_size, PROT_READ | PROT_WRITE,
		      MAP_SHARED, vtfd, 0);
	if (window == MAP_FAILED) {
		/*
		 * ioctl(vtfd, KDDISABIO, 0);
		 */
		ioctl(vtfd, FBIO_SETMODE, &prev_mode);
		ioctl(vtfd, VT_SETMODE, &prev_vtmode);
		err(1, "failed to mmap the video buffer");
	}

	/*
	 * Load palette, clear the frame buffer, and draw something.
	 */
	if (setjmp(&loop) == 0) {
		in_main_loop = TRUE;
		while (!quit) {
			draw_something();
		}
		in_main_loop = FALSE;
	}

	/* Closing down our shop... */
	munmap(window, win_size);

	/*
	 * End direct access to the video card registers (see above).
	 *
	 * ioctl(vtfd, KDDISABIO, 0);
	 */
	ioctl(vtfd, FBIO_SETMODE, &prev_mode);	/* SW_XXX */
	ioctl(vtfd, VT_SETMODE, &prev_vtmode);

	exit(0);
}

static void
draw_something(void)
{
	u_char *p;
	int pos;
	int x, y;
	u_char val;

	...

	/* 
	 * Calculate the position of the pixel as the offset from the 
	 * beginning of the frame buffer 
	 */
	pos = line_width*y + x*bpp/planes/8;

	/* If bpp < 8, we need to calculate bit offset and bit mask too... */

	/* 
	 * Change the widow origin so that the bank containing the above
	 * pixel will appear in the window.
	 * Note that this will always work even if the video card is in
	 * the linear mode.
	 */
	ioctl(vtfd, FBIO_SETWINORG, &pos);	/* CONS_SETWINORG */

	/* 
	 * Form a pointer by adding the offset within the window to
	 * the window address.
	 * Note that the following calculation is valid both in
	 * the linear mode and the window mode.
	 */
	p = window + pos%win_size;

	/* read/write or do whatever you like to the pixel */
	val = *p;
	*p = 0;

	...
}

static void
sig_release(int sig)
{
	/* Always refuse to release this vty */
	ioctl(vtfd, VT_RELDISP, VT_FALSE);
}

static void
sig_acquire(int sig)
{
	/* Always acknowledge that we are acquiring the vty. */
	ioctl(vtfd, VT_RELDISP, VT_ACKACQ);
}

static void
sig_term(int sig)
{
	quit = TRUE;
	signal(SIGINT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	if (in_main_loop)
		longjmp(loop, 1);
}
