/*-
 * fbaccess3.c
 *
 * A sample code for frame buffer access.
 * Kazutaka YOKOTA.
 */

#include <stdio.h>
#include <err.h>
#include <fcntl.h>
#include <signal.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fbio.h>
#include <sys/consio.h>

#define TRUE	1
#define FALSE	0

/* Our virtual terminal. */
static int vtfd;

/* Our frame buffer */
static int fbfd;

/* Set to TRUE, if the program is to terminate. */
static int quit = FALSE;

/* The original video mode. */
static int prev_mode;

/* The frame buffer window. */
static u_char *window = MAP_FAILED;
static int win_size = 0;		/* window size */
static int buf_size = 0;		/* frame buffer size */
static int line_width;
static int bpp;				/* bits per pixel */
static int planes;			/* color planes */

/* Function prototypes */
static void draw_something(void);
static void sig_term(int);

int
main(int argc, char *argv[])
{
	video_adapter_info_t adp_info;
	video_info_t mode_info;
	char *fbname;
	int adapter;
	int mode;

	/* Use stdin. */
	vtfd = 0;

	/* 
	 * Open an arbitrary frame buffer.  If the frame buffer is 
	 * associated with vtys by the console driver, the following
	 * call to open() will fail.
	 */
	fbname = "/dev/fb1";
	fbfd = open(fbname, O_RDWR);
	if (fbfd < 0)
		err(1, "cannot open %s", fbname);
	ioctl(fbfd, FBIO_ADAPTER, &adapter);

	/* Desired video mode */
	mode = M_VESA_CG1024x768;

	/*
	 * Obtain video mode information.  If this ioctl fails, the video
	 * card does not support the specified mode.
	 */
	mode_info.vi_mode = mode;
	if (ioctl(fbfd, FBIO_MODEINFO, &mode_info))
		err(1, "video mode %d is not supported", mode);

	/* Save the current video mode */
	ioctl(fbfd, FBIO_GETMODE, &prev_mode);

	/* Set up signal handlers. */
	signal(SIGINT, sig_term);
	signal(SIGTERM, sig_term);

	/* Set our video mode */
	ioctl(fbfd, FBIO_SETMODE, &mode);

	/*
	 * Obtain the adapter state and information.  It is essential to
	 * call this ioctl to examine exactly how the video card is set up.
	 * Window size and buffer size parameters must be obtained by 
	 * FBIO_ADPINFO. The corresponding values obtained by FBIO_MODEINFO
	 * shouldn't be used. 
	 */
	ioctl(fbfd, FBIO_ADPINFO, &adp_info);
	win_size = adp_info.va_window_size;
	buf_size = adp_info.va_buffer_size;
	line_width = adp_info.va_line_width;

	/* The other information is available in mode_info.  */
	bpp = mode_info.vi_depth;
	planes = mode_info.vi_planes;

	/*
	 * If win_size < buf_size/planes, the video card is in the window 
	 * mode. The program must switch memory banks to access the desired
	 * part of the frame buffer. 
	 */

	/* Now everything is set... */

	/*
	 * Map the window/frame buffer to the process address space.
	 */
	window = mmap(0, win_size, PROT_READ | PROT_WRITE,
		      MAP_SHARED, fbfd, 0);
	if (window == MAP_FAILED) {
		ioctl(fbfd, FBIO_SETMODE, &prev_mode);
		close(fbfd);
		err(1, "unable to mmap the video buffer");
	}

	/*
	 * Load palette, clear the frame buffer, and draw something.
	 */
	while (!quit) {
		draw_something();
	}

	/* Closing down our shop... */
	munmap(window, win_size);
	ioctl(fbfd, FBIO_SETMODE, &prev_mode);	/* SW_XXX */
	close(fbfd);

	exit(0);
}

static void
draw_something(void)
{
	u_char *p;
	int pos;
	int x, y;
	u_char val;

	...

	/* 
	 * Calculate the position of the pixel as the offset from the 
	 * beginning of the frame buffer 
	 */
	pos = line_width*y + x*bpp/planes/8;

	/* If bpp < 8, we need to calculate bit offset and bit mask too... */

	/* 
	 * Change the widow origin so that the bank containing the above
	 * pixel will appear in the window.
	 * Note that this will always work even if the video card is in
	 * the linear mode.
	 */
	ioctl(fbfd, FBIO_SETWINORG, &pos);

	/* 
	 * Form a pointer by adding the offset within the window to
	 * the window address.
	 * Note that the following calculation is valid both in
	 * the linear mode and the window mode.
	 */
	p = window + pos%win_size;

	/* read/write or do whatever you like to the pixel */
	val = *p;
	*p = 0;

	...
}

static void
sig_term(int sig)
{
	signal(SIGINT, SIG_IGN);
	signal(SIGTERM, SIG_IGN);

	if (window != MAP_FAILED) {
		munmap(window, win_size);
		window = MAP_FAILED;
	}
	ioctl(fbfd, FBIO_SETMODE, &prev_mode);	/* SW_XXX */
	close(fbfd);

	exit(1);
}
